//
//  ac.cu
//  VecTripper
//
//  Created by tmkk on 09/07/31.
//  Copyright 2009 tmkk. All rights reserved.
//

#include "ac.h"

#include <stdlib.h>
#include <string.h>

typedef struct _queue
{
	struct _queue *next;
	struct _queue *last;
	AC_node *data;
} AC_queue;

static void queue(AC_node *node, AC_queue *root)
{
	AC_queue *q = (AC_queue *)malloc(sizeof(AC_queue));
	q->next = NULL;
	q->data = node;
	root->last->next = q;
	root->last = q;
}

static AC_node *dequeue(AC_queue *root)
{
	if(!root->next) return NULL;
	AC_node *ret = root->next->data;
	AC_queue *newnode = root->next->next;
	free(root->next);
	root->next = newnode;
	if(!root->next) root->last = root;
	return ret;
}

static void free_queue(AC_queue *root)
{
	if(root->next) free_queue(root->next);
	free(root);
}

static int AC_new_node(AC_t *ac)
{
	ac->trie = (AC_node*)realloc(ac->trie,sizeof(AC_node)*(++ac->nodenum+1));
	memset(NODEPTR(ac->nodenum,ac), 0, sizeof(AC_node));
	return ac->nodenum;
}

AC_t *AC_new(void)
{
	AC_t *ac = (AC_t*)calloc(sizeof(AC_t), 1);
	ac->nodenum = 1;
	ac->trie = (AC_node*)calloc(sizeof(AC_node)*2, 1);
	ac->minlen = 12;
	return ac;
}

void AC_add_pattern(char *pattern, int length, int idx, AC_t *ac)
{
	int i;
	AC_node *node = NODEPTR(1,ac);
	int oldnode = 1;
	for (i=0;i<length;i++) {
		int c = pattern[i];
		if(node->next[c] == 0) {
			int newnode = AC_new_node(ac);
			node = NODEPTR(oldnode,ac);
			node->next[c] = newnode;
			if(c<32) node->valid1 |= 1 << c;
			else node->valid2 |= 1 << (c-32);
			NODEPTR(newnode,ac)->distance = i+1;
		}
		oldnode = node->next[c];
		node = NODEPTR(oldnode,ac);
    }
	if(idx < 0) for(i=0;i<12;i++) node->idx[i] = 1;
	else {
		int pos = idx+length-1;
		if(pos >= 0 && pos < 12) node->idx[pos] = 1;
	}
	if(ac->minlen > length) ac->minlen = length;
}

void AC_init_tree(AC_t *ac)
{
	int i,c;
	AC_node *root = NODEPTR(1,ac);
	AC_node *node, *node2;
	AC_queue *Q = (AC_queue *)malloc(sizeof(AC_queue));
	Q->next = NULL;
	Q->last = Q;
	for(c=0;c<64;c++) {
		if(root->next[c]) {
			NODEPTR(root->next[c],ac)->next[64] = 1;
			for(i=0;i<64;i++) {
				if(!NODEPTR(root->next[c],ac)->next[i]) {
					NODEPTR(root->next[c],ac)->next[i] = 1;
				}
			}
			queue(NODEPTR(root->next[c],ac), Q);
		}
		else root->next[c] = 1;
	}
	while(Q->next) {
		node = dequeue(Q);
		for(c=0;c<64;c++) {
			if(c<32 ? (node->valid1&(1<<c)) : (node->valid2&(1<<(c-32)))) {
				queue(NODEPTR(node->next[c],ac), Q);
				node2 = NODEPTR(node->next[64],ac);
				while(!node2->next[c]) node2 = NODEPTR(node2->next[64],ac);
				NODEPTR(node->next[c],ac)->next[64] = node2->next[c];
				for(i=0;i<12;i++) {
					if(NODEPTR(node2->next[c],ac)->idx[i]) NODEPTR(node->next[c],ac)->idx[i] = 1;
				}
				for(i=0;i<64;i++) {
					if(!NODEPTR(node->next[c],ac)->next[i]) {
						node2 = NODEPTR(NODEPTR(node->next[c],ac)->next[64],ac);
						while(!node2->next[i]) node2 = NODEPTR(node2->next[64],ac);
						NODEPTR(node->next[c],ac)->next[i] = node2->next[i];
					}
				}
			}
		}
	}
	free_queue(Q);
}

void AC_free(AC_t *ac)
{
	free(ac->trie);
	free(ac);
}
