//
//  main.cu
//  VecTripper
//
//  Created by tmkk on 09/07/31.
//  Copyright 2009 tmkk. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <iconv.h>
#include <getopt.h>
#include <cutil.h>
#include "ac.h"

#define NUM_THREAD 160
#define NUM_THREAD_G200 256

typedef struct
{
	unsigned int IN[3];
	unsigned int OUT[3];
} sha1_cuda_t;

__constant__ int gpu_validchar[128];
//__constant__ char gpu_pat[13];
__constant__ unsigned int gpu_pat[3];
__constant__ unsigned int gpu_mask[3];
__constant__ AC_node gpu_ac[100];

static const char base64[65] ="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789./";

static unsigned int rand_x=123456789,rand_y=362436069,rand_z=521288629,rand_w=88675123;

#define AC_MATCH(i) \
node = &gpu_ac[node->next[c]]; \
distance = node->distance; \
limit = 12-pat_len+distance; \
if(distance < pat_len) { \
	if(i >= limit) goto nextloop; \
} \
else if(node->idx[i]) { \
	sha1_in->IN[0] = out[0]; \
	sha1_in->IN[1] = out[1]; \
	sha1_in->IN[2] = out[2]; \
	sha1_in->OUT[0] = in[0]; \
	sha1_in->OUT[1] = in[1]; \
	sha1_in->OUT[2] = in[2]; \
	goto nextloop; \
} \
else if(i >= limit) goto nextloop;

#include "sha1core.cu"

__global__ void sha1_cuda(sha1_cuda_t *input, int pat_len, int mode)
{
	int state;
	unsigned int in[3], out[3];
	char trip[12];
	AC_node *node;
	sha1_cuda_t *sha1_in = &input[blockIdx.x * blockDim.x + threadIdx.x];
	in[0] = sha1_in->IN[0];
	in[1] = sha1_in->IN[1];
	in[2] = sha1_in->IN[2];
	
	for(state=0;state<0x3fff;state++) {
		in[2] = (in[2] & 0xffff0000) | (gpu_validchar[(state >> 7) & 0x7f] << 8) | (gpu_validchar[state & 0x7f]);
		sha1core_cuda(in, out);
		
		if(!(((out[1] >> 22) & 0x3f)^((out[1] >> 16) & 0x3f))) {
			trip[0] = (out[0] >> 26);
			trip[1] = (out[0] >> 20) & 0x3f;
			trip[2] = (out[0] >> 14) & 0x3f;
			trip[3] = (out[0] >> 8) & 0x3f;
			trip[4] = (out[0] >> 2) & 0x3f;
			trip[5] = ((out[0] & 0x3) << 4)|(out[1] >> 28);
			trip[6] = (out[1] >> 22) & 0x3f;
			trip[7] = (out[1] >> 16) & 0x3f;
			trip[8] = (out[1] >> 10) & 0x3f;
			trip[9] = (out[1] >> 4) & 0x3f;
			trip[10] = ((out[1] & 0xf) << 2)|(out[2] >> 30);
			trip[11] = (out[2] >> 24) & 0x3f;
			int cont = 1;
			if(!(trip[1]^trip[0])) cont++;
			if(!(trip[2]^trip[1])) cont++;
			else cont = 1;
			if(!(trip[3]^trip[2])) cont++;
			else cont = 1;
			if(!(trip[4]^trip[3])) cont++;
			else goto next;
			if(!(trip[5]^trip[4])) cont++;
			else goto next;
			if(!(trip[6]^trip[5])) cont++;
			else goto next;
			if(!(trip[7]^trip[6])) cont++;
			else goto next;
			if(!(trip[8]^trip[7])) cont++;
			else goto end;
			if(!(trip[9]^trip[8])) cont++;
			else goto end;
			if(!(trip[10]^trip[9])) cont++;
			else goto end;
			if(!(trip[11]^trip[10])) cont++;
		  end:
			if(cont > 7) {
				sha1_in->IN[0] = out[0];
				sha1_in->IN[1] = out[1];
				sha1_in->IN[2] = out[2];
				sha1_in->OUT[0] = in[0];
				sha1_in->OUT[1] = in[1];
				sha1_in->OUT[2] = in[2];
			}
		}
	  next:
		switch(mode) {
		  case 0: /* partial match */
			int c, distance, limit;
			node = gpu_ac+1;
			
			c = (out[0] >> 26);
			AC_MATCH(0);
			c = (out[0] >> 20) & 0x3f;
			AC_MATCH(1);
			c = (out[0] >> 14) & 0x3f;
			AC_MATCH(2);
			c = (out[0] >> 8) & 0x3f;
			AC_MATCH(3);
			c = (out[0] >> 2) & 0x3f;
			AC_MATCH(4);
			c = ((out[0] & 0x3) << 4)|(out[1] >> 28);
			AC_MATCH(5);
			c = (out[1] >> 22) & 0x3f;
			AC_MATCH(6);
			c = (out[1] >> 16) & 0x3f;
			AC_MATCH(7);
			c = (out[1] >> 10) & 0x3f;
			AC_MATCH(8);
			c = (out[1] >> 4) & 0x3f;
			AC_MATCH(9);
			c = ((out[1] & 0xf) << 2)|(out[2] >> 30);
			AC_MATCH(10);
			c = (out[2] >> 24) & 0x3f;
			AC_MATCH(11);
			
			break;
#if 0
		  case 1: /* begin match */
			for(i=0;i<pat_len;i++) {
				if(gpu_pat[i] != trip[i]) goto nextloop;
			}
			sha1_in->IN[0] = out[0];
			sha1_in->IN[1] = out[1];
			sha1_in->IN[2] = out[2];
			sha1_in->OUT[0] = in[0];
			sha1_in->OUT[1] = in[1];
			sha1_in->OUT[2] = in[2];
			break;
		  case 2: /* last match */
			for(i=0;i<pat_len;i++) {
				if(gpu_pat[i] != trip[12-pat_len+i]) goto nextloop;
			}
			sha1_in->IN[0] = out[0];
			sha1_in->IN[1] = out[1];
			sha1_in->IN[2] = out[2];
			sha1_in->OUT[0] = in[0];
			sha1_in->OUT[1] = in[1];
			sha1_in->OUT[2] = in[2];
			break;
#else
		  default:
			if((out[0]&gpu_mask[0]) != gpu_pat[0]) goto nextloop;
			if((out[1]&gpu_mask[1]) != gpu_pat[1]) goto nextloop;
			if((out[2]&gpu_mask[2]) == gpu_pat[2]) {
				sha1_in->IN[0] = out[0];
				sha1_in->IN[1] = out[1];
				sha1_in->IN[2] = out[2];
				sha1_in->OUT[0] = in[0];
				sha1_in->OUT[1] = in[1];
				sha1_in->OUT[2] = in[2];
			}
#endif
		}
	  nextloop:
	}
}

static unsigned int xor128(){
	unsigned int t;
	t=(rand_x^(rand_x<<11));rand_x=rand_y;rand_y=rand_z;rand_z=rand_w; return( rand_w=(rand_w^(rand_w>>19))^(t^(t>>8)) );
}

static int base64dec(char c)
{
	if(c >= 'A' && c <= 'Z') return c - 'A';
	else if(c >= 'a' && c <= 'z') return c - 'a' + 26;
	else if(c >= '0' && c <= '9') return c - '0' + 52;
	else if(c == '.') return 62;
	else if(c == '/') return 63;
	else return 0;
}

static void generate_mask(char *str, unsigned int *masks)
{
	int i;
	unsigned int mask, val;
	/* 1-5 */
	mask = 0;
	for(i=0;i<5;i++) {
		val = str[i];
		mask |= (val << (26-i*6));
	}
	/* 6 */
	val = str[5];
	mask |= (val >> 4);
	*masks = mask;
	mask = 0;
	mask |= ((val & 0xf) << 28);
	/* 7-10 */
	for(i=0;i<4;i++) {
		val = str[i+6];
		mask |= (val << (22-i*6));
	}
	/* 11 */
	val = str[10];
	mask |= (val >> 2);
	*(masks+1) = mask;
	mask = 0;
	mask |= ((val & 0x3) << 30);
	/* 12 */
	val = str[11];
	mask |= (val << 24);
	*(masks+2) = mask;
}

static void generate_pattern(char *str, unsigned int *pats)
{
	int i;
	unsigned int pat, val;
	/* 1-5 */
	pat = 0;
	for(i=0;i<5;i++) {
		val = str[i];
		pat |= (val << (26-i*6));
	}
	/* 6 */
	val = str[5];
	pat |= (val >> 4);
	*pats = pat;
	pat = 0;
	pat |= ((val & 0xf) << 28);
	/* 7-10 */
	for(i=0;i<4;i++) {
		val = str[i+6];
		pat |= (val << (22-i*6));
	}
	/* 11 */
	val = str[10];
	pat |= (val >> 2);
	*(pats+1) = pat;
	pat = 0;
	pat |= ((val & 0x3) << 30);
	/* 12 */
	val = str[11];
	pat |= (val << 24);
	*(pats+2) = pat;
}

static void usage(char *argv[])
{
	fprintf(stderr,"Usage: %s [-d deviceNum] <target str>\n\t%s Apple\n\t%s \"^Apple\"\n\t%s \"Apple$\"\n",argv[0],argv[0],argv[0],argv[0]);
}

int main(int argc, char *argv[])
{
	int i,j;
	int validchar[128];
	char pat[12];
	int pat_len, pat_idx, mode;
	int opt, device=0;
	struct timeval tv, tv2;
	iconv_t cd;
	
	if(argc < 2) {
		usage(argv);
		return 0;
	}
	
	while ((opt = getopt(argc, argv, "d:")) != -1) {
		switch (opt) {
		  case 'd':
			device = atoi(optarg);
			break;
		  default:
			usage(argv);
			return 0;
		}
	}
	
	if(optind >= argc) {
		usage(argv);
		return 0;
	}
	
	int deviceCount;
	CUDA_SAFE_CALL_NO_SYNC(cudaGetDeviceCount(&deviceCount));
	if (deviceCount == 0) {
		fprintf(stderr, "Error: no devices supporting CUDA found.\n");
		return 0;
	}
	
	//CUT_DEVICE_INIT(1,argv);
	
	cudaDeviceProp deviceProp;
	cudaGetDeviceProperties(&deviceProp, device);
	
	if(deviceProp.major < 1) {
		fprintf(stderr, "Error: device does not support CUDA.\n");
		return 0;
	}
	
	CUDA_SAFE_CALL(cudaSetDevice(device));
	
	int num_block = deviceProp.multiProcessorCount * 8;
	int num_thread = deviceProp.regsPerBlock<10000 ? NUM_THREAD : NUM_THREAD_G200;
	
	fprintf(stderr,"Using device %d: %s (%d SPs)\n", device, deviceProp.name, deviceProp.multiProcessorCount*8);
	
	dim3 threads(num_thread,1,1); 
	dim3 grid(num_block,1,1);
	
	cd = iconv_open("UTF-8","SHIFT-JIS");
	
	gettimeofday(&tv, NULL);
	rand_x = tv.tv_sec ^ tv.tv_usec;
	
	for(i=37,j=0;i<127;i++) { // 90
		validchar[j++] = i;
	}
	for(i=161;i<199;i++) { // 38
		validchar[j++] = i;
	}
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(gpu_validchar, validchar, sizeof(int)*128));
	
	pat_len = strlen(argv[optind]);
	if(argv[optind][0] == '^') {
		mode = 1;
		pat_idx = 1;
		pat_len -= 1;
	}
	else if(argv[optind][pat_len-1] == '$') {
		mode = 2;
		pat_idx = 0;
		pat_len -= 1;
	}
	else {
		mode = 0;
		pat_idx = 0;
	}
	
	if(pat_len < 5) {
		fprintf(stderr, "Error: pattern length is too short! at least 5 characters are required.\n");
		return 0;
	}
	else if(pat_len > 12) {
		fprintf(stderr, "Error: pattern length is too long!\n");
		return 0;
	}
	
	for(i=0;i<pat_len;i++) {
		pat[i] = base64dec(argv[optind][i+pat_idx]);
	}
	
	if(mode == 0) {
		AC_t *ac = AC_new();
		AC_add_pattern(pat, pat_len, -1, ac);
		AC_init_tree(ac);
		//CUDA_SAFE_CALL(cudaMalloc(reinterpret_cast<void**>(&gpu_ac), (ac->nodenum+1)*sizeof(AC_node)));
		CUDA_SAFE_CALL(cudaMemcpyToSymbol(gpu_ac, ac->trie, (ac->nodenum+1)*sizeof(AC_node)));
		AC_free(ac);
	}
	else {
		unsigned int p[3],m[3];
		if(mode == 1) {
			for(i=pat_len;i<12;i++) pat[i] = 0;
			generate_pattern(pat,p);
			for(i=0;i<pat_len;i++) pat[i] = 0x3f;
			generate_mask(pat,m);
		}
		else {
			for(i=pat_len-1;i>=0;i--) pat[i+12-pat_len] = pat[i];
			for(i=0;i<12-pat_len;i++) pat[i] = 0;
			generate_pattern(pat,p);
			for(i=12-pat_len;i<12;i++) pat[i] = 0x3f;
			generate_mask(pat,m);
		}
		CUDA_SAFE_CALL(cudaMemcpyToSymbol(gpu_pat, p, 12));
		CUDA_SAFE_CALL(cudaMemcpyToSymbol(gpu_mask, m, 12));
	}
	
	int dataSize = sizeof(sha1_cuda_t) * num_thread * num_block;
	sha1_cuda_t* in = (sha1_cuda_t *)calloc(dataSize, 1);
	sha1_cuda_t* gpu_in;
	CUDA_SAFE_CALL(cudaMalloc(reinterpret_cast<void**>(&gpu_in), dataSize));

	gettimeofday(&tv, NULL);
	while(1) {
		
		for(j=0;j<num_thread*num_block;j++) {
			unsigned int randNum = xor128();
			in[j].IN[0] = (validchar[(randNum >> 24) & 0x7f] << 24) | (validchar[(randNum >> 16) & 0x7f] << 16) | (validchar[(randNum >> 8) & 0x7f] << 8) | (validchar[randNum & 0x7f]);
			randNum = xor128();
			in[j].IN[1] = (validchar[(randNum >> 24) & 0x7f] << 24) | (validchar[(randNum >> 16) & 0x7f] << 16) | (validchar[(randNum >> 8) & 0x7f] << 8) | (validchar[randNum & 0x7f]);
			randNum = xor128();
			in[j].IN[2] = (validchar[(randNum >> 8) & 0x7f] << 24) | (validchar[(randNum) & 0x7f] << 16);
			
			in[j].OUT[0] = 0;
			in[j].OUT[1] = 0;
			in[j].OUT[2] = 0;
		}
		CUDA_SAFE_CALL(cudaMemcpy(gpu_in, in, dataSize , cudaMemcpyHostToDevice));
		
		sha1_cuda<<<grid,threads>>>(gpu_in, pat_len, mode);
		
		CUDA_SAFE_CALL(cudaMemcpy(in, gpu_in, dataSize, cudaMemcpyDeviceToHost));
		
		int first = 1;
		for(j=0;j<num_thread*num_block;j++) {
			if(in[j].OUT[0] != 0) {
				
				if(first) {
					printf("\n");
					first = 0;
				}
				char trip[13];
				char key[13],key_out[32];
				key[12] = trip[12] = 0;
				trip[0] = base64[(in[j].IN[0] >> 26)];
				trip[1] = base64[(in[j].IN[0] >> 20) & 0x3f];
				trip[2] = base64[(in[j].IN[0] >> 14) & 0x3f];
				trip[3] = base64[(in[j].IN[0] >> 8) & 0x3f];
				trip[4] = base64[(in[j].IN[0] >> 2) & 0x3f];
				trip[5] = base64[((in[j].IN[0] & 0x3) << 4)|(in[j].IN[1] >> 28)];
				trip[6] = base64[(in[j].IN[1] >> 22) & 0x3f];
				trip[7] = base64[(in[j].IN[1] >> 16) & 0x3f];
				trip[8] = base64[(in[j].IN[1] >> 10) & 0x3f];
				trip[9] = base64[(in[j].IN[1] >> 4) & 0x3f];
				trip[10] = base64[((in[j].IN[1] & 0xf) << 2)|(in[j].IN[2] >> 30)];
				trip[11] = base64[(in[j].IN[2] >> 24) & 0x3f];
				key[0] = in[j].OUT[0] >> 24;
				key[1] = (in[j].OUT[0] >> 16) & 0xff;
				key[2] = (in[j].OUT[0] >> 8) & 0xff;
				key[3] = (in[j].OUT[0]) & 0xff;
				key[4] = in[j].OUT[1] >> 24;
				key[5] = (in[j].OUT[1] >> 16) & 0xff;
				key[6] = (in[j].OUT[1] >> 8) & 0xff;
				key[7] = (in[j].OUT[1]) & 0xff;
				key[8] = in[j].OUT[2] >> 24;
				key[9] = (in[j].OUT[2] >> 16) & 0xff;
				key[10] = (in[j].OUT[2] >> 8) & 0xff;
				key[11] = (in[j].OUT[2]) & 0xff;
				
				char *iconv_in = key;
				char *iconv_out = key_out;
				size_t in_len = 12;
				size_t out_len = 31;
				out_len = iconv(cd, &iconv_in, &in_len, &iconv_out, &out_len);
				*iconv_out = 0;
				printf("<>%s : #%s\n",trip,key_out);
				FILE *fp = fopen("out.txt", "ab");
				fprintf(fp,"<>%s : #%s\n",trip,key_out);
				fclose(fp);
			}
		}
		
		gettimeofday(&tv2, NULL);
		float dur = (tv2.tv_sec-tv.tv_sec)*1e+6+tv2.tv_usec-tv.tv_usec;
		tv = tv2;
		fprintf(stderr,"\rSpeed: %f Mtrips/s",(float)0x3fff*num_thread*num_block/dur);
		fflush(stdout);
	}
	iconv_close(cd);
	cudaFree(gpu_in);
	free(in);

	CUT_EXIT(argc, argv);

	return 0;
}
