//
//  sha1core.cu
//  VecTripper
//
//  Created by tmkk on 09/07/31.
//  Copyright 2009 tmkk. All rights reserved.
//

#define H0 0x67452301;
#define H1 0xEFCDAB89;
#define H2 0x98BADCFE;
#define H3 0x10325476;
#define H4 0xC3D2E1F0;
#define K0 0x5A827999;
#define K1 0x6ED9EBA1;
#define K2 0x8F1BBCDC;
#define K3 0xCA62C1D6;

#define ROTL(val, amt) ((val)<<(amt) | (val)>>(32-amt))
#define F1(b, c, d) ((d)^((b)&((c)^(d))))
#define F2(b, c, d) ((b)^(c)^(d))
#define F3(b, c, d) (((b)&((c)|(d)))|((c)&(d)))


#define ROUND1(IN1, IN2, IN3, IN4, IN5, W1, stage) \
W1 = in[stage]; \
v1 = ROTL(IN1, 5); \
v2 = F1(IN2, IN3, IN4); \
v3 = W1 + K0; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND1a(IN1, IN2, IN3, IN4, IN5, W1, stage) \
W1 = 0x80000000; \
v1 = ROTL(IN1, 5); \
v2 = F1(IN2, IN3, IN4); \
v3 = W1 + K0; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND1b(IN1, IN2, IN3, IN4, IN5, W1, stage) \
W1 = 0x00000060; \
v1 = ROTL(IN1, 5); \
v2 = F1(IN2, IN3, IN4); \
v3 = W1 + K0; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND1c(IN1, IN2, IN3, IN4, IN5, W1, stage) \
W1 = 0; \
v1 = ROTL(IN1, 5); \
v2 = F1(IN2, IN3, IN4); \
IN5 += v1 + v2 + K0; \
IN2 = ROTL(IN2, 30);

#define ROUND1d(IN1, IN2, IN3, IN4, IN5, W1, W2, W3, W4, stage) \
W1 = W1 ^ W2; \
W1 = ROTL(W1, 1); \
v1 = ROTL(IN1, 5); \
v2 = F1(IN2, IN3, IN4); \
v3 = W1 + K0; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND1e(IN1, IN2, IN3, IN4, IN5, W1, W2, W3, W4, stage) \
W1 = W1 ^ W4; \
W1 = ROTL(W1, 1); \
v1 = ROTL(IN1, 5); \
v2 = F1(IN2, IN3, IN4); \
v3 = W1 + K0; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND2(IN1, IN2, IN3, IN4, IN5, W1, W2, W3, W4, stage) \
W1 = W1 ^ W2 ^ W3 ^ W4; \
W1 = ROTL(W1, 1); \
v1 = ROTL(IN1, 5); \
v2 = F2(IN2, IN3, IN4); \
v3 = W1 + K1; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND2b(IN1, IN2, IN3, IN4, IN5, W1, W2, W3, W4, stage) \
W1 = ROTL(W4, 1); \
v1 = ROTL(IN1, 5); \
v2 = F2(IN2, IN3, IN4); \
v3 = W1 + K1; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND2c(IN1, IN2, IN3, IN4, IN5, W1, W2, W3, W4, stage) \
W1 = W3 ^ W4; \
W1 = ROTL(W1, 1); \
v1 = ROTL(IN1, 5); \
v2 = F2(IN2, IN3, IN4); \
v3 = W1 + K1; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND3(IN1, IN2, IN3, IN4, IN5, W1, W2, W3, W4, stage) \
W1 = W1 ^ W2 ^ W3 ^ W4; \
W1 = ROTL(W1, 1); \
v1 = ROTL(IN1, 5); \
v2 = F3(IN2, IN3, IN4); \
v3 = W1 + K2; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

#define ROUND4(IN1, IN2, IN3, IN4, IN5, W1, W2, W3, W4, stage) \
W1 = W1 ^ W2 ^ W3 ^ W4; \
W1 = ROTL(W1, 1); \
v1 = ROTL(IN1, 5); \
v2 = F2(IN2, IN3, IN4); \
v3 = W1 + K3; \
IN5 += v1 + v2 + v3; \
IN2 = ROTL(IN2, 30);

__device__ void sha1core_cuda(unsigned int *in, unsigned int *out)
{
	unsigned int v1, v2, v3;
	unsigned int vA, vB, vC, vD, vE;
	unsigned int vW1, vW2, vW3, vW4, vW5, vW6, vW7, vW8, vW9, vW10, vW11, vW12, vW13, vW14, vW15, vW16;

	vA = H0;
	vB = H1;
	vC = H2;
	vD = H3;
	vE = H4;
	
	ROUND1(vA, vB, vC, vD, vE, vW1, 0);
	ROUND1(vE, vA, vB, vC, vD, vW2, 1);
	ROUND1(vD, vE, vA, vB, vC, vW3, 2);
	ROUND1a(vC, vD, vE, vA, vB, vW4, 3);
	ROUND1c(vB, vC, vD, vE, vA, vW5, 4);
	ROUND1c(vA, vB, vC, vD, vE, vW6, 5);
	ROUND1c(vE, vA, vB, vC, vD, vW7, 6);
	ROUND1c(vD, vE, vA, vB, vC, vW8, 7);
	ROUND1c(vC, vD, vE, vA, vB, vW9, 8);
	ROUND1c(vB, vC, vD, vE, vA, vW10, 9);
	ROUND1c(vA, vB, vC, vD, vE, vW11, 10);
	ROUND1c(vE, vA, vB, vC, vD, vW12, 11);
	ROUND1c(vD, vE, vA, vB, vC, vW13, 12);
	ROUND1c(vC, vD, vE, vA, vB, vW14, 13);
	ROUND1c(vB, vC, vD, vE, vA, vW15, 14);
	ROUND1b(vA, vB, vC, vD, vE, vW16, 15);
	ROUND1d(vE, vA, vB, vC, vD, vW1, vW3, vW9, vW14, 16);
	ROUND1d(vD, vE, vA, vB, vC, vW2, vW4, vW10, vW15, 17);
	ROUND1e(vC, vD, vE, vA, vB, vW3, vW5, vW11, vW16, 18);
	ROUND1e(vB, vC, vD, vE, vA, vW4, vW6, vW12, vW1, 19);

	ROUND2b(vA, vB, vC, vD, vE, vW5, vW7, vW13, vW2, 20);
	ROUND2b(vE, vA, vB, vC, vD, vW6, vW8, vW14, vW3, 21);
	ROUND2b(vD, vE, vA, vB, vC, vW7, vW9, vW15, vW4, 22);
	ROUND2c(vC, vD, vE, vA, vB, vW8, vW10, vW16, vW5, 23);
	ROUND2c(vB, vC, vD, vE, vA, vW9, vW11, vW1, vW6, 24);
	ROUND2c(vA, vB, vC, vD, vE, vW10, vW12, vW2, vW7, 25);
	ROUND2c(vE, vA, vB, vC, vD, vW11, vW13, vW3, vW8, 26);
	ROUND2c(vD, vE, vA, vB, vC, vW12, vW14, vW4, vW9, 27);
	ROUND2c(vC, vD, vE, vA, vB, vW13, vW15, vW5, vW10, 28);
	ROUND2(vB, vC, vD, vE, vA, vW14, vW16, vW6, vW11, 29);
	ROUND2(vA, vB, vC, vD, vE, vW15, vW1, vW7, vW12, 30);
	ROUND2(vE, vA, vB, vC, vD, vW16, vW2, vW8, vW13, 31);
	ROUND2(vD, vE, vA, vB, vC, vW1, vW3, vW9, vW14, 32);
	ROUND2(vC, vD, vE, vA, vB, vW2, vW4, vW10, vW15, 33);
	ROUND2(vB, vC, vD, vE, vA, vW3, vW5, vW11, vW16, 34);
	ROUND2(vA, vB, vC, vD, vE, vW4, vW6, vW12, vW1, 35);
	ROUND2(vE, vA, vB, vC, vD, vW5, vW7, vW13, vW2, 36);
	ROUND2(vD, vE, vA, vB, vC, vW6, vW8, vW14, vW3, 37);
	ROUND2(vC, vD, vE, vA, vB, vW7, vW9, vW15, vW4, 38);
	ROUND2(vB, vC, vD, vE, vA, vW8, vW10, vW16, vW5, 39);

	ROUND3(vA, vB, vC, vD, vE, vW9, vW11, vW1, vW6, 40);
	ROUND3(vE, vA, vB, vC, vD, vW10, vW12, vW2, vW7, 41);
	ROUND3(vD, vE, vA, vB, vC, vW11, vW13, vW3, vW8, 42);
	ROUND3(vC, vD, vE, vA, vB, vW12, vW14, vW4, vW9, 43);
	ROUND3(vB, vC, vD, vE, vA, vW13, vW15, vW5, vW10, 44);
	ROUND3(vA, vB, vC, vD, vE, vW14, vW16, vW6, vW11, 45);
	ROUND3(vE, vA, vB, vC, vD, vW15, vW1, vW7, vW12, 46);
	ROUND3(vD, vE, vA, vB, vC, vW16, vW2, vW8, vW13, 47);
	ROUND3(vC, vD, vE, vA, vB, vW1, vW3, vW9, vW14, 48);
	ROUND3(vB, vC, vD, vE, vA, vW2, vW4, vW10, vW15, 49);
	ROUND3(vA, vB, vC, vD, vE, vW3, vW5, vW11, vW16, 50);
	ROUND3(vE, vA, vB, vC, vD, vW4, vW6, vW12, vW1, 51);
	ROUND3(vD, vE, vA, vB, vC, vW5, vW7, vW13, vW2, 52);
	ROUND3(vC, vD, vE, vA, vB, vW6, vW8, vW14, vW3, 53);
	ROUND3(vB, vC, vD, vE, vA, vW7, vW9, vW15, vW4, 54);
	ROUND3(vA, vB, vC, vD, vE, vW8, vW10, vW16, vW5, 55);
	ROUND3(vE, vA, vB, vC, vD, vW9, vW11, vW1, vW6, 56);
	ROUND3(vD, vE, vA, vB, vC, vW10, vW12, vW2, vW7, 57);
	ROUND3(vC, vD, vE, vA, vB, vW11, vW13, vW3, vW8, 58);
	ROUND3(vB, vC, vD, vE, vA, vW12, vW14, vW4, vW9, 59);

	ROUND4(vA, vB, vC, vD, vE, vW13, vW15, vW5, vW10, 60);
	ROUND4(vE, vA, vB, vC, vD, vW14, vW16, vW6, vW11, 61);
	ROUND4(vD, vE, vA, vB, vC, vW15, vW1, vW7, vW12, 62);
	ROUND4(vC, vD, vE, vA, vB, vW16, vW2, vW8, vW13, 63);
	ROUND4(vB, vC, vD, vE, vA, vW1, vW3, vW9, vW14, 64);
	ROUND4(vA, vB, vC, vD, vE, vW2, vW4, vW10, vW15, 65);
	ROUND4(vE, vA, vB, vC, vD, vW3, vW5, vW11, vW16, 66);
	ROUND4(vD, vE, vA, vB, vC, vW4, vW6, vW12, vW1, 67);
	ROUND4(vC, vD, vE, vA, vB, vW5, vW7, vW13, vW2, 68);
	ROUND4(vB, vC, vD, vE, vA, vW6, vW8, vW14, vW3, 69);
	ROUND4(vA, vB, vC, vD, vE, vW7, vW9, vW15, vW4, 70);
	ROUND4(vE, vA, vB, vC, vD, vW8, vW10, vW16, vW5, 71);
	ROUND4(vD, vE, vA, vB, vC, vW9, vW11, vW1, vW6, 72);
	ROUND4(vC, vD, vE, vA, vB, vW10, vW12, vW2, vW7, 73);
	ROUND4(vB, vC, vD, vE, vA, vW11, vW13, vW3, vW8, 74);
	ROUND4(vA, vB, vC, vD, vE, vW12, vW14, vW4, vW9, 75);
	ROUND4(vE, vA, vB, vC, vD, vW13, vW15, vW5, vW10, 76);
	ROUND4(vD, vE, vA, vB, vC, vW14, vW16, vW6, vW11, 77);
	ROUND4(vC, vD, vE, vA, vB, vW15, vW1, vW7, vW12, 78);
	ROUND4(vB, vC, vD, vE, vA, vW16, vW2, vW8, vW13, 79);

	out[0] = vA + H0;
	out[1] = vB + H1;
	out[2] = vC + H2;
	__syncthreads();
}
